"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.SchemaPropertyType = void 0;
var SchemaLiteral_1 = require("./SchemaLiteral");
var SchemaPropertyType = /** @class */ (function () {
    function SchemaPropertyType(typeName, types) {
        this._isNullable = typeName.endsWith("?");
        if (this._isNullable) {
            typeName = typeName.substr(0, typeName.length - 1);
        }
        this._isArray = typeName.endsWith("[]");
        if (this._isArray) {
            typeName = typeName.substr(0, typeName.length - 2);
        }
        this._isDictionary = typeName.startsWith("Dictionary<") && typeName.endsWith(">");
        if (this._isDictionary) {
            typeName = typeName.substr("Dictionary<".length);
            typeName = typeName.substr(0, typeName.length - 1);
        }
        switch (typeName) {
            case "uri":
                this._type = SchemaLiteral_1.SchemaLiteral.uri;
                break;
            case "uri-reference":
                this._type = SchemaLiteral_1.SchemaLiteral.uriReference;
                break;
            case "string":
                this._type = SchemaLiteral_1.SchemaLiteral.string;
                break;
            case "number":
                this._type = SchemaLiteral_1.SchemaLiteral.number;
                break;
            case "boolean":
                this._type = SchemaLiteral_1.SchemaLiteral.boolean;
                break;
            case "object":
                this._type = SchemaLiteral_1.SchemaLiteral.object;
                break;
            case "any":
                this._type = SchemaLiteral_1.SchemaLiteral.any;
                break;
            default:
                // Must be an object reference
                var foundType = types.get(typeName);
                if (foundType == undefined) {
                    throw new Error("Property references " + typeName + " type, but " + typeName + " couldn't be found.");
                }
                this._type = foundType;
                break;
        }
    }
    Object.defineProperty(SchemaPropertyType.prototype, "type", {
        get: function () {
            return this._type;
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(SchemaPropertyType.prototype, "isNullable", {
        get: function () {
            return this._isNullable;
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(SchemaPropertyType.prototype, "isArray", {
        get: function () {
            return this._isArray;
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(SchemaPropertyType.prototype, "isDictionary", {
        get: function () {
            return this._isDictionary;
        },
        enumerable: false,
        configurable: true
    });
    return SchemaPropertyType;
}());
exports.SchemaPropertyType = SchemaPropertyType;
//# sourceMappingURL=SchemaPropertyType.js.map