"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __exportStar = (this && this.__exportStar) || function(m, exports) {
    for (var p in m) if (p !== "default" && !exports.hasOwnProperty(p)) __createBinding(exports, m, p);
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.transformTypes = exports.transformFolder = exports.markdown = void 0;
var SchemaClass_1 = require("./SchemaClass");
var SchemaEnum_1 = require("./SchemaEnum");
var SchemaLiteral_1 = require("./SchemaLiteral");
var Schema_1 = require("./Schema");
var markdown = require("./markdown/index");
exports.markdown = markdown;
var fs = require('fs');
var path = require('path');
__exportStar(require("./SchemaType"), exports);
__exportStar(require("./SchemaClass"), exports);
__exportStar(require("./SchemaEnum"), exports);
__exportStar(require("./SchemaProperty"), exports);
__exportStar(require("./SchemaPropertyType"), exports);
__exportStar(require("./SchemaLiteral"), exports);
__exportStar(require("./Schema"), exports);
__exportStar(require("./TransformOptions"), exports);
__exportStar(require("./markdown/index"), exports);
function transformFolder(pathToTypeFiles, options) {
    return new Transformer(Schema_1.Schema.fromFolder(pathToTypeFiles), options).transform();
}
exports.transformFolder = transformFolder;
function transformTypes(types, options) {
    return new Transformer(new Schema_1.Schema(types), options).transform();
}
exports.transformTypes = transformTypes;
function isObjectEmpty(obj) {
    if (!obj) {
        return true;
    }
    for (var key in obj) {
        return false;
    }
    return true;
}
var Transformer = /** @class */ (function () {
    function Transformer(schema, options) {
        var _this = this;
        this._primaryTypes = [];
        this._definitions = {};
        this._implementationsOf = {};
        this._extendables = {};
        this._schema = schema;
        this._typePropertyName = options.typePropertyName || 'type';
        this._defaultPrimaryTypeName = options.defaultPrimaryTypeName;
        this._allowAdditionalProperties = options.allowAdditionalProperties;
        this._allowCustomEnums = options.allowCustomEnums;
        this._allowCustomTypes = options.allowCustomTypes;
        this._enforceEnumCaseSensitivity = options.enforceEnumCaseSensitivity;
        if (options.primaryTypeNames) {
            options.primaryTypeNames.forEach(function (value) {
                _this._primaryTypes.push(_this.getType(value));
            });
        }
    }
    Transformer.prototype.transform = function () {
        var _this = this;
        // First create all definitions for all types
        this._schema.typeDictionary.forEach(function (type) {
            _this.defineType(type);
        });
        var answer = {
            $schema: 'http://json-schema.org/draft-06/schema#',
            id: '',
            definitions: {}
        };
        var anyOf = [];
        this._primaryTypes.forEach(function (primaryType) {
            var definition = {
                allOf: [
                    {
                        $ref: '#/definitions/' + primaryType.type
                    }
                ]
            };
            // If there's multiple primary types, we have to enforce type at top level to differentiate
            if (_this._primaryTypes.length > 1) {
                // If it's not the default primary type name, make type required
                if (primaryType.type !== _this._defaultPrimaryTypeName) {
                    definition.required = [_this._typePropertyName];
                }
            }
            anyOf.push(definition);
        });
        answer.anyOf = anyOf;
        if (!isObjectEmpty(this._definitions)) {
            answer.definitions = __assign({}, this._definitions);
        }
        if (!isObjectEmpty(this._implementationsOf)) {
            for (var key in this._implementationsOf) {
                var anyOfValue = [];
                this._implementationsOf[key].forEach(function (implementationTypeName) {
                    anyOfValue.push({
                        required: [_this._typePropertyName],
                        allOf: [
                            {
                                $ref: '#/definitions/' + implementationTypeName
                            }
                        ]
                    });
                });
                answer.definitions['ImplementationsOf.' + key] = {
                    anyOf: anyOfValue
                };
            }
        }
        if (!isObjectEmpty(this._extendables)) {
            for (var key in this._extendables) {
                answer.definitions['Extendable.' + key] = this._extendables[key];
            }
        }
        return answer;
    };
    Transformer.prototype.getType = function (typeName) {
        return this._schema.getType(typeName);
    };
    Transformer.prototype.transformType = function (type) {
        var _this = this;
        try {
            if (type instanceof SchemaEnum_1.SchemaEnum) {
                var enumType = type;
                var transformed = __assign({}, type.original);
                delete transformed.type;
                delete transformed.values;
                delete transformed.classType;
                delete transformed.$schema;
                delete transformed.extends;
                var enums = [];
                //增加原始枚举值包括description
                var originalEnums = [];
                enumType.values.forEach(function (val) {
                    enums.push(val.value);
                    originalEnums.push(val.original);
                });
                // If only allows strict enums but case-insensitive...
                if (!this._enforceEnumCaseSensitivity && !this._allowCustomEnums) {
                    var enumRegexs = [];
                    enums.forEach(function (enumVal) {
                        enumRegexs.push('(' + _this.toCaseInsensitiveRegex(enumVal) + ')');
                    });
                    var regex = '^' + enumRegexs.join('|') + '$';
                    transformed.anyOf = [
                        {
                            enum: enums,
                            originEnum: originalEnums
                        },
                        {
                            pattern: regex
                        }
                    ];
                }
                else if (!this._allowCustomEnums) {
                    // Else strict and sensitive
                    transformed.enum = enums;
                    transformed.originEnum = originalEnums;
                }
                else {
                    // Else allows any enums, so no need for the regex
                    transformed.anyOf = [
                        {
                            enum: enums,
                            originEnum: originalEnums
                        },
                        {
                            type: 'string'
                        }
                    ];
                }
                return transformed;
            }
            else if (!(type instanceof SchemaClass_1.SchemaClass)) {
                throw new Error('Unknown class type ' + type);
            }
            var transformed = __assign({}, type.original);
            transformed.type = 'object';
            if (!this._allowAdditionalProperties) {
                transformed.additionalProperties = false;
            }
            delete transformed.isAbstract;
            delete transformed.classType;
            delete transformed.$schema;
            delete transformed.extends;
            delete transformed.shorthand;
            transformed.properties = {};
            var overridesProperties = false;
            var overriddenProperties = [];
            var properties = new Map();
            if (type.extends.length > 0) {
                type.properties.forEach(function (prop) {
                    if (prop.override) {
                        // If property exists in base classes
                        if (type.getAllExtendedProperties().indexOf(prop.name) !== -1) {
                            // Note that JSON schema doesn't support the concept of overriding properties, so instead
                            // we need to copy all the properties and NOT inherit.
                            overridesProperties = true;
                            overriddenProperties.push(prop.name);
                        }
                        else {
                            console.warn(type.type + "." + prop.name + " has override set to true, but it doesn't override any base class properties.");
                        }
                    }
                    else {
                        // If property exists in base classes
                        if (type.getAllExtendedProperties().indexOf(prop.name) !== -1) {
                            console.warn("Ignoring " + type.type + "." + prop.name + " as there's already this property specified in a base class. If you want to keep this property, add \"override\": \"true\".");
                        }
                    }
                });
            }
            if (overridesProperties) {
                type.getAllProperties().forEach(function (propVal, key) {
                    properties.set(key, propVal);
                });
            }
            else {
                properties = type.properties;
            }
            if (properties.size > 0) {
                properties.forEach(function (propVal, key) {
                    var transformedPropVal;
                    try {
                        transformedPropVal = _this.transformPropertyValue(propVal);
                    }
                    catch (err) {
                        throw new Error('Failed transforming property ' + key + '\n\n' + err.stack);
                    }
                    if (transformedPropVal.required) {
                        delete transformedPropVal.required;
                        if (!transformed.required) {
                            transformed.required = [];
                        }
                        transformed.required.push(key);
                    }
                    transformed.properties[key] = transformedPropVal;
                });
            }
            if (!type.isAbstract) {
                // If it's not abstract, we add the type property
                // Note that we don't require it though, it's optional
                // var newProperties: any = {};
                // newProperties[this._typePropertyName] = {
                // 	"enum": [type.type],
                // 	"description": "Must be `" + type.type + "`"
                // };
                transformed.properties = __assign({}, transformed.properties);
            }
            if (type.extends.length > 0) {
                if (!overridesProperties) {
                    // Have to add placeholders for all the properties
                    type.getAllExtendedProperties().forEach(function (extendedPropKey) {
                        // If there's an existing property defined, skip
                        var existingProp = type.properties.get(extendedPropKey);
                        if (existingProp !== undefined) {
                            if (!existingProp.override) {
                                console.warn("Overriding extended property " + extendedPropKey + " on type " + type.type + ". If this was intentional, add \"override\": \"true\" to this property to prevent this warning from appearing.");
                            }
                            return;
                        }
                        transformed.properties[extendedPropKey] = {};
                    });
                    transformed.allOf = [];
                    type.extends.forEach(function (extended) {
                        if (extended.getAllProperties().size > 0) {
                            transformed.allOf.push({
                                $ref: '#/definitions/Extendable.' + extended.type
                            });
                        }
                    });
                    if (transformed.allOf.length == 0) {
                        delete transformed.allOf;
                    }
                }
                transformed.parentRefs = [];
                type.extends.forEach(function (extended) {
                    transformed.parentRefs.push({
                        $ref: '#/definitions/Extendable.' + extended.type
                    });
                });
                if (overriddenProperties && overriddenProperties.length) {
                    transformed.overriddenProperties = overriddenProperties;
                }
                // Keep track of implementations
                type.getAllExtended().forEach(function (extended) {
                    if (!_this._implementationsOf[extended.type]) {
                        _this._implementationsOf[extended.type] = [];
                        // If extending type isn't abstract, add that as an implementation
                        if (!extended.isAbstract) {
                            _this._implementationsOf[extended.type].push(extended.type);
                        }
                    }
                    if (!type.isAbstract) {
                        _this._implementationsOf[extended.type].push(type.type);
                    }
                });
                delete transformed.extends;
            }
            if (type.shorthand) {
                transformed.anyOf = [
                    __assign({}, transformed.properties[type.shorthand.name]),
                    {
                        type: transformed.type,
                        properties: transformed.properties,
                        required: transformed.required,
                        additionalProperties: false
                    }
                ];
                delete transformed.type;
                delete transformed.properties;
                delete transformed.required;
                delete transformed.additionalProperties;
            }
            return transformed;
        }
        catch (err) {
            throw 'Failed transforming type ' + type.type + '\n\n' + err.stack;
        }
    };
    Transformer.prototype.toCaseInsensitiveRegex = function (enumValue) {
        var answer = '';
        for (var i = 0; i < enumValue.length; i++) {
            var c = enumValue.charAt(i);
            answer += "[" + c.toLowerCase() + "|" + c.toUpperCase() + "]";
        }
        return answer;
    };
    Transformer.prototype.transformPropertyValue = function (propertyValue) {
        var _this = this;
        var cleanPropertyValue = __assign({}, propertyValue.original);
        delete cleanPropertyValue.type;
        delete cleanPropertyValue.shorthands;
        delete cleanPropertyValue.override;
        var types = [];
        var values = [];
        propertyValue.types.forEach(function (type) {
            types.push(type);
        });
        propertyValue.shorthands.forEach(function (shorthand) {
            shorthand.types.forEach(function (type) {
                types.push(type);
            });
        });
        types.forEach(function (type) {
            var transformedValue = {};
            if (type.type instanceof SchemaLiteral_1.SchemaLiteral) {
                if (type.type.schemaType) {
                    transformedValue.type = type.type.schemaType;
                }
                if (type.type.schemaFormat) {
                    transformedValue.format = type.type.schemaFormat;
                }
            }
            else {
                // Must be an object reference
                transformedValue.$ref =
                    '#/definitions/' +
                        (_this.hasMultipleImplementations(type.type.type) ? 'ImplementationsOf.' : '') +
                        type.type.type;
            }
            if (type.isDictionary) {
                transformedValue = {
                    type: 'object',
                    additionalProperties: __assign({}, transformedValue)
                };
            }
            if (type.isArray) {
                transformedValue = {
                    type: 'array',
                    items: __assign({}, transformedValue)
                };
                if (type.type == SchemaLiteral_1.SchemaLiteral.any) {
                    // Any types don't need type definitions
                    delete transformedValue.items;
                }
            }
            propertyValue.shorthands.forEach(function (shorthand) {
                if (shorthand.types.indexOf(type) != -1) {
                    var cleanShorthandValue = __assign({}, shorthand.original);
                    delete cleanShorthandValue.type;
                    transformedValue = __assign(__assign({}, transformedValue), cleanShorthandValue);
                }
            });
            values.push(transformedValue);
            if (type.isNullable) {
                values.push({
                    type: 'null'
                });
            }
        });
        if (values.length == 1) {
            return __assign(__assign({}, values[0]), cleanPropertyValue);
        }
        else {
            return __assign({ anyOf: values }, cleanPropertyValue);
        }
    };
    Transformer.prototype.hasMultipleImplementations = function (typeName) {
        var type = this._schema.typeDictionary.get(typeName);
        if (type instanceof SchemaClass_1.SchemaClass && type.implementations.length > 0) {
            return true;
        }
        return false;
    };
    Transformer.prototype.defineType = function (type) {
        var typeName = type.type;
        var transformedType = this.transformType(type);
        // If there's multiple implementations of this type
        if (this.hasMultipleImplementations(typeName) && type instanceof SchemaClass_1.SchemaClass && type.getAllProperties().size > 0) {
            // Then we must define an extendable flavor that doesn't have the additionalProperties = false
            var extendableType = __assign({}, transformedType);
            delete extendableType.additionalProperties;
            this._extendables[typeName] = extendableType;
        }
        // As long as it's not abstract (no reason to define those, they'll be defined as extendable if used)
        if (!(type instanceof SchemaClass_1.SchemaClass && type.isAbstract)) {
            this._definitions[typeName] = transformedType;
        }
    };
    return Transformer;
}());
//# sourceMappingURL=typed-schema.js.map